// //See package.json for details


// //in theory we can let ReadWithHighlighting do the work and just expose these as public functions
// exports.SetContainerId = function(id) {
//     ReadWithHighlighting.SetContainerId(id); //no return
// };
// exports.HasSelection = function() {
//     return ReadWithHighlighting.HasSelection(); //returns true or false
// };
// exports.SelectCurrentSentence = function() {
//     return ReadWithHighlighting.SelectCurrentSentence(); //returns false if no sentence could be selected
// };
// exports.GetSelection = function() {
//    return ReadWithHighlighting.GetSelection(); //returns the selection as a range
// };
// exports.SelectNextSentence = function() {
//     return ReadWithHighlighting.SelectNextSentence(); //returns false if the next sentence couldn't be highlighted
// };
// exports.CollapseSelection = function() {
//     ReadWithHighlighting.CollapseSelection(); //no return
// };

  /*
	ReadWithHighlighting 
	Library for making an element (assumed to have contenteditable true) allow selecting and highlighting
	and iterating through its contents.
	
	USAGE
	Call ReadWithHighlighting.SetContainerId with the id of the element that you are using to display content/reading.
	Check ReadWithHighlighting.HasSelection. If you have a selection, then (I assume) you won't want to read through the text. Call
		GetSelection to get it. 
	Call ReadWithHighlighting.SelectCurrentSentence. If you get true back, then this has worked.
	Call ReadWithHighlighting.GetSelection to get the text of the sentence to read.
	Call ReadWithHighlighting.SelectNextSentence to progress. If this returns false, you've finished.
	Call ReadWithHighlighting.CollapseSelection to put the caret at the end when you've finished speaking.
	
	Alasdair, October 2014.
*/

var ReadWithHighlighting = {
	RichTextContainerId : ""
};

ReadWithHighlighting.SetContainerId = function (id) {
	ReadWithHighlighting.RichTextContainerId = id;
};

ReadWithHighlighting.HasSelection = function () {
	var r = document.getSelection().getRangeAt(0);
	return !r.collapsed;
};

ReadWithHighlighting.GetSelection = function () {
	var r = document.getSelection().getRangeAt(0);
	return "" + r;
};

ReadWithHighlighting.SelectNextSentence = function () {
	ReadWithHighlighting.CollapseSelection();
	return ReadWithHighlighting.SelectCurrentSentence();
};

ReadWithHighlighting.CollapseSelection = function () {
	var sel = window.getSelection();
	if (sel && !sel.isCollapsed){
		var r = sel.getRangeAt(0); 
		r.collapse(false); // argument required in IE
		sel.removeAllRanges();
		sel.addRange(r);
	}
};

ReadWithHighlighting.startsWithCapital = function (str) {
	return (str[0] === str[0].toUpperCase());
};

ReadWithHighlighting.SelectFirstSentence = function(){
	var treeWalker, firstTextNode = null, sel, range;

	treeWalker = document.createTreeWalker(
		document.body, // root node for the traversal
		NodeFilter.SHOW_TEXT | NodeFilter.SHOW_ELEMENT, // a set of constants to filter against
		null,
		false
	);
	try {
		while (treeWalker.nextNode()) { // Can throw an error!
			if (treeWalker.currentNode.nodeName === "#text"){
				firstTextNode = treeWalker.currentNode;
				break;
			}
		}
	} catch (err) { return false; }

	if (firstTextNode){
		range = document.createRange();
		sel = document.getSelection();
		range.setStart(firstTextNode, 0);
		range.setEnd(firstTextNode, 0);
		sel.removeAllRanges();
		sel.addRange(range);

		return this.SelectCurrentSentence(range);
	}
};

ReadWithHighlighting.SelectCurrentSentence = function (r = null, d = null) {
	var newStart, newEnd, treeWalker, foundCurrent, found, delimiter, nodeList, i, lastTextNode, goneForward = false, capitalLetterText = false, doc = document;
	if (d !== null){
		doc = d;
	}

	if (ReadWithHighlighting.RichTextContainerId) {
		doc.body.focus();
	}

	if (r === null){
		if (document && doc.getSelection && doc.getSelection().rangeCount > 0 && doc.getSelection().getRangeAt){
			r = doc.getSelection().getRangeAt(0);
		}
	}

	if (!r)
		return false;
	if (!r.startContainer)
		return false; // no selection
	if (r.startContainer.nodeName === "BODY")
		return false;
	if (r.startContainer.nodeName === "HTML")
		return false;
	if (r.startContainer.nodeName === "#document")
		return false;
	
	//console.log("scs1 Find back");
	if (r.collapsed) {
		if (!ReadWithHighlighting.GotoNextTextNode(doc)) {
			//console.log("Failed to go to next TextNode");
			return false;
		}
		r = doc.getSelection().getRangeAt(0); 
		
		// Select back:
		newStart = ReadWithHighlighting.FindSentenceDelimiter(r.startContainer, r.startOffset, -1, goneForward);
		capitalLetterText = ReadWithHighlighting.startsWithCapital(r.startContainer.nodeValue);
		if (newStart !== -1) {
			// OK, all the text we need is in the current node.
			newStart = (newStart > 0 || (newStart === 0 && "!.?".indexOf(r.startContainer.nodeValue.charAt(0) === 0))) ? newStart + 1 : newStart;
			r.setStart(r.startContainer, newStart);	
		} else if (capitalLetterText){
			// No sentence delimiter in this node, but it starts with a capital letter so assume its the start of a sentence
			r.setStart(r.startContainer, 0);
		} else {
			// Oh, we have to iterate back through the tree until we get to a block node or a 
			// sentence start.
			treeWalker = doc.createTreeWalker(
				doc.body, // root node for the traversal
				NodeFilter.SHOW_TEXT | NodeFilter.SHOW_ELEMENT, // a set of constants to filter against
				null,
				false
			);
			nodeList = [];
			try {
				while (treeWalker.nextNode()) { // Can throw an error!
					if (treeWalker.currentNode === r.startContainer) {
						break;
					} else {
						nodeList.push(treeWalker.currentNode);
					}
				}
			} catch (err) { return false; }
			// We now have an array of nodes before the node with the caret. Go back, looking for 
			// a block element or a sentence delimiter.
			newStart = -1;
			lastTextNode = r.startContainer;
			found = false;

			for (i = nodeList.length - 1; i > -1; i--) {
				if (nodeList[i].nodeName === "#text") {
					newStart = ReadWithHighlighting.FindSentenceDelimiterFromEnd(nodeList[i], "end", goneForward);
					capitalLetterText = ReadWithHighlighting.startsWithCapital(nodeList[i].nodeValue);
					if (newStart !== -1) {
						// Ah, got it!
						r.setStart(nodeList[i], newStart+1);
						found = true;
						break;
					} else if (capitalLetterText){
						r.setStart(nodeList[i], 0);
						found = true;
						break;
					}
					lastTextNode = nodeList[i];
				} else if (ReadWithHighlighting.IsBlockNode(nodeList[i])) {
					// Ah, we've crashed into a block! Just select from the start of the last 
					// text node we found.
					r.setStart(lastTextNode, 0);
					found = true;
					break;
				}
			}
			// Did we find a start?
			if (!found) {
				// Ah, well, just go with the last text node.
				r.setStart(lastTextNode, 0);
			}
		}
		
		//console.log("scs2 Find forwards. Text at this point: [" + r + "]");
		// We're in a text node, which has some text: at the very least, we're going to read that.
		newEnd = ReadWithHighlighting.FindSentenceDelimiter(r.endContainer, r.endOffset, 1, goneForward);
		if (newEnd !== -1) {
			// OK! all the text we need is in the current node.
			if ((newEnd + 1) > r.endContainer.nodeValue.length){
				newEnd = r.endContainer.nodeValue.length;
			}
			r.setEnd(r.endContainer, newEnd + 1);
			// And we're done.
			//console.log("scs3 Sentence end in this text node.");
			ReadWithHighlighting.Select(r);
			return true;
		} else {
			goneForward = true;
			// Need to progress on through the tree, until we either get to a sentence delimiter
			// or a block node.
			treeWalker = doc.createTreeWalker(
				doc.body, // root node for the traversal
				NodeFilter.SHOW_TEXT | NodeFilter.SHOW_ELEMENT, // a set of constants to filter against
				null,
				false
			);
			foundCurrent = false;
			found = null;
			delimiter = -1;
			try {
				var forwardOnce = false;
				while (treeWalker.nextNode()) {
					if (treeWalker.currentNode === r.endContainer) {
						foundCurrent = true;
						//we're at the caret position now, we already determined there is no delimiter in the rest of this node, so jump forward one before continuing.
						//but only do this the first time otherwise we'll miss out every other node.
						if (!forwardOnce){
							treeWalker.nextNode();
							forwardOnce = true;
						}
					}
					if (foundCurrent && treeWalker.currentNode.nodeName === "#text") {
						delimiter = ReadWithHighlighting.FindSentenceDelimiterFromEnd(treeWalker.currentNode, "START", goneForward);
						if (delimiter !== -1) {
							// Yay!
							delimiter = (delimiter > 0) ? delimiter + 1 : delimiter;
							r.setEnd(treeWalker.currentNode, delimiter);
							ReadWithHighlighting.Select(r);
							return true;
						} else {
							//console.log("scs13 Expanded to encompass: " + treeWalker.currentNode.nodeValue);
							r.setEnd(treeWalker.currentNode, treeWalker.currentNode.nodeValue.length);
						}
					} /*else if (foundCurrent && ReadWithHighlighting.IsBlockNode(treeWalker.currentNode)) {
						// Stop! Done.
						ReadWithHighlighting.Select(r);
						return true;
					}*/
				}
			} catch (err) { return false; }
		} 
	} else {
		// Got a selection
		return false;
	}
};

// From https://dvcs.w3.org/hg/editing/raw-file/tip/editing.html#issues = list of nodes with layout.
//A prohibited paragraph child name is "address", "article", "aside", "blockquote", "caption", "center", "col", "colgroup", "dd", "details", "dir", "div", "dl", "dt", "fieldset", "figcaption", "figure", "footer", "form", "h1", "h2", "h3", "h4", "h5", "h6", "header", "hgroup", "hr", "li", "listing", "menu", "nav", "ol", "p", "plaintext", "pre", "section", "summary", "table", "tbody", "td", "tfoot", "th", "thead", "tr", "ul", or "xmp".
// Added also BR so that this indicates a "block" ending - that is, a linebreak, which means reading stops. 
ReadWithHighlighting.IsBlockNode = function (node) {
	if (!node)
		return false;
	return ("*ADDRESS*ARTICLE*ASIDE*BLOCKQUOTE*BR*CAPTION*CENTER*COL*COLGROUP*DD*DETAILS*DIR*DIV*DL*DT*FIELDSET*FIGCAPTION*FIGURE*FOOTER*FORM*H1*H2*H3*H4*H5*H6*HEADER*HGROUP*HR*LI*LISTING*MENU*NAV*OL*P*PLAINTEXT*PRE*SECTION*SUMMARY*TABLE*TBODY*TD*TFOOT*TH*THEAD*TR*UL*XMP*".indexOf("*" + node.nodeName + "*") !== -1);		
};

ReadWithHighlighting.FindSentenceDelimiterFromEnd = function (textNode, fromWhichEnd, goneForward) {
	if (!textNode)
		return -1;
	if (!textNode.nodeName) 
		return -1;
	if (textNode.nodeName != "#text") 
		return -1;
	if (fromWhichEnd && fromWhichEnd.toUpperCase() === "END") {
		return ReadWithHighlighting.FindSentenceDelimiter(textNode, textNode.nodeValue.length - 1, -1, goneForward);
	} else {
		return ReadWithHighlighting.FindSentenceDelimiter(textNode, 0, 1, goneForward);
	}
};

ReadWithHighlighting.GotoNextTextNode = function (doc) {
	var r = doc.getSelection().getRangeAt(0),
		treeWalker, foundCurrent, found; 
	if (!r.collapsed) 
		return false;
	if (r.startContainer.nodeName === "#text") {
		var s = r.startContainer.nodeValue;
		s = s.substring(r.startOffset);
		if (ReadWithHighlighting.HasValidText(s)) {
			return r.startContainer;
		} else {
			// We're in a text node, but no valid text. No good.
		}
	}
	// Proceed through the rest of the tree, returning the first valid text node.
	treeWalker = doc.createTreeWalker(
		doc.body, // root node for the traversal
		NodeFilter.SHOW_TEXT, // a set of constants to filter against
		null,
		false
	);
	foundCurrent = false;
	found = null;
	while (treeWalker.nextNode()) {
		if (foundCurrent) {
			if (ReadWithHighlighting.HasValidText(treeWalker.currentNode.nodeValue)) {
				found = treeWalker.currentNode;
				break;
			}
		}
		if (treeWalker.currentNode === r.startContainer) {
			foundCurrent = true;
		}
	}
	if (found) {
		r = doc.getSelection().getRangeAt(0);
		r.setStart(found, 0);
		ReadWithHighlighting.Select(r);
		return true;
	}
	return false;
};

// Does everything necessary to select and display the range r, including scroll into view.
ReadWithHighlighting.Select = function (r) {
	var sel = window.getSelection();
	sel.removeAllRanges();
	sel.addRange(r); // Need to re-select (for WebKit)
	//r.startContainer.parentNode.scrollIntoView(true);
};

ReadWithHighlighting.HasValidText = function (stringToTest) {
	var s = stringToTest.replace(/ /g, "");
	s = s.replace(/\t/g, "");
	s = s.replace(/\n/g, "");
	s = s.replace(/\r/g, "");
	s = s.replace(/\f/g, "");
	return (s.length > 0);
};

// Find a sentence delimiter - ".?!:" in the text node, starting at startIndex and progressing in the
// direction of direction. Return the index of the delimiter, or -1 if not found. 
// If the character at the startIndex is a sentence delimiter, this will count.
ReadWithHighlighting.FindSentenceDelimiter = function (textNode, startIndex, direction, goneForward) {
	var s, i, c, cNext;
	if (!textNode)
		return -1;
	if (!textNode.nodeName)
		return -1;
	if (textNode.nodeName != "#text")
		return -1;
	if (startIndex < 0)
		return -1;
	s = textNode.nodeValue;
	if (s.length === 0)
		return -1;
	if (startIndex > s.length - 1)
		return -1;
		
	if ((direction != -1) && (direction != 1))
		return -1;

	for (i = startIndex; i > -1 && i < s.length; i += direction) {
		c = s.charAt(i);
		if ("!.?".indexOf(c) !== -1) {
			// Found a punctuation mark! Check if it whitespace or end of node follows immediately after.
			if (i < (s.length - 1)){
				cNext = s.charCodeAt(i+1);
				if (cNext === 32 || cNext === 13 || cNext === 10){
					//punctuation is followed by whitespace, this is the end of the sentence.
					return i;
				}
			} else {
				//punctuation at end of node, assume end of sentence.
				return i;
			}
		}
		// if current character is a capital or number, and previous character is a new line, assume this is a new sentence.
		// or if current character is a capital or number and there is no previous character in this node, also assume this is a new sentence.
		if ((c.charCodeAt(0) <= 90 && c.charCodeAt(0) >= 65) || (c.charCodeAt(0) <= 57 && c.charCodeAt(0) >= 48)){
			//got a capital letter or number
			var pc = s.charAt(i-1);
			if (pc.charCodeAt(0) === 10) {
				return i-1;
			} else if (pc.charCodeAt(0) === 13 || (i === 0 && goneForward)){
				return i;
			}
		}
	}
	// Nope.
	return -1;
};